/*
 * Sonar, entreprise quality control tool.
 * Copyright (C) 2007-2008 Hortis-GRC SA
 * mailto:be_agile HAT hortis DOT ch
 *
 * Sonar is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * Sonar is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with Sonar; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02
 */
package org.sonar.plugins.surefire;

import ch.hortis.sonar.model.Metric;
import ch.hortis.sonar.service.MeasureKey;
import org.apache.commons.lang.StringUtils;
import org.apache.maven.plugin.logging.Log;
import org.sonar.plugins.api.maven.AbstractJavaMavenCollector;
import org.sonar.plugins.api.maven.CollectsUnitTests;
import org.sonar.plugins.api.maven.JavaMeasuresRecorder;
import org.sonar.plugins.api.maven.MavenCollectorUtils;
import org.sonar.plugins.api.maven.MavenPluginHandler;
import org.sonar.plugins.api.maven.model.MavenPom;
import org.sonar.plugins.api.maven.xml.XmlParserException;
import org.sonar.plugins.api.metrics.CoreMetrics;
import org.w3c.dom.Document;
import org.w3c.dom.Node;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;
import java.io.File;
import java.io.FilenameFilter;
import java.text.ParseException;

public class SurefireMavenCollector extends AbstractJavaMavenCollector implements CollectsUnitTests {
  private String overridenReportsDir=null;

  public SurefireMavenCollector(JavaMeasuresRecorder measuresRecorder, Log log) {
    super(measuresRecorder, log);
  }

  @Override
  public boolean shouldCollectIfLightMode() {
    return false;
  }

  public boolean shouldCollectIfNoSources() {
    return true;
  }

  public Class<? extends MavenPluginHandler> dependsOnMavenPlugin(MavenPom pom) {
    return SurefireMavenPluginHandler.class;
  }

  public boolean shouldStopOnFailure() {
    return false;
  }

  public void overrideReportsDirectory(String s) {
    overridenReportsDir=s;
  }

  public void collect(MavenPom pom) {
    collect(getSurefireReportsDir(pom));
  }

  private File getSurefireReportsDir(MavenPom pom) {
    if (overridenReportsDir==null) {
      return new File(pom.getBuildDirectory() + "/surefire-reports");
    }
    return new File(pom.getBuildDirectory() + overridenReportsDir);
  }

  protected void collect(File surefireDir) {
    File[] reports = findSureFireReportFiles(surefireDir);
    if (reports.length == 0) {
      getMeasuresRecorder().createProjectMeasure(new MeasureKey(CoreMetrics.TESTS_COUNT), 0.0);

    } else {
      try {

        DocumentBuilder builder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
        for (File report : reports) {
          Document document = builder.parse(report);
          XPath xpath = XPathFactory.newInstance().newXPath();
          UnitTestReport fileReport = parseReport(xpath, document);

          saveClassMeasure(fileReport, CoreMetrics.TESTS_SKIPPED, fileReport.skipped);
          saveClassMeasure(fileReport, CoreMetrics.TESTS_COUNT, fileReport.tests);
          saveClassMeasure(fileReport, CoreMetrics.TESTS_ERRORS, fileReport.errors);
          saveClassMeasure(fileReport, CoreMetrics.TESTS_FAILURES, fileReport.failures);
          if (!Double.isNaN(fileReport.time)) {
            saveClassMeasure(fileReport, CoreMetrics.TESTS_TIME, getMeasureTime(fileReport));
          }
        }

      } catch (Exception e) {
        throw new XmlParserException("parsing Surefire xml files from " + surefireDir, e);
      }
    }
  }

  private void saveClassMeasure(UnitTestReport fileReport, Metric metric, double value) {
    if (!Double.isNaN(value)) {
      getMeasuresRecorder().createClassMeasure(fileReport.packageName, fileReport.className, true, new MeasureKey(metric), value);
    }
  }

  private double getMeasureTime(UnitTestReport fileReport) {
    return MavenCollectorUtils.scaleValue(fileReport.time * 1000, 3);
  }

  private UnitTestReport parseReport(XPath xpath, Node root) throws XPathExpressionException, ParseException {
    Node testSuite = (Node) xpath.evaluate("//testsuite", root, XPathConstants.NODE);
    UnitTestReport test = new UnitTestReport();
    String fullname = xpath.evaluate("@name", testSuite);
    test.packageName = StringUtils.substringBeforeLast(fullname, ".");
    test.className = StringUtils.substringAfterLast(fullname, ".");
    test.errors = MavenCollectorUtils.parseNumber(xpath.evaluate("@errors", testSuite));
    test.skipped = MavenCollectorUtils.parseNumber(xpath.evaluate("@skipped", testSuite));
    test.tests = MavenCollectorUtils.parseNumber(xpath.evaluate("@tests", testSuite));
    test.time = MavenCollectorUtils.parseNumber(xpath.evaluate("@time", testSuite));
    test.failures = MavenCollectorUtils.parseNumber(xpath.evaluate("@failures", testSuite));
    return test;
  }

  private File[] findSureFireReportFiles(File sureFireOutputDirectory) {
    if (sureFireOutputDirectory == null || !sureFireOutputDirectory.exists()) {
      return new File[0];
    }

    return sureFireOutputDirectory.listFiles(new FilenameFilter() {
      public boolean accept(File dir, String name) {
        return name.startsWith("TEST-") && name.endsWith(".xml");
      }
    });
  }

  private class UnitTestReport {
    private String packageName;
    private String className;
    private double errors;
    private double skipped;
    private double tests;
    private double time;
    private double failures;
  }
}
